@@ -2,14 +2,20 @@ package Geo::ShapeFile::Shape;
 use strict;
 use warnings;
 use Carp;
+use Tree::R;
+use List::Util qw /min max/;
+
 use Geo::ShapeFile;
 use Geo::ShapeFile::Point;
+use Geo::ShapeFile::Shape::Index;
 
 use parent qw /Geo::ShapeFile/;
-our $VERSION = '2.54';
 
-my $is_little_endian = unpack 'b', (pack 'S', 1 );
+our $VERSION = '2.56';
 
+my $little_endian_sys = unpack 'b', (pack 'S', 1 );
+
+my $index_class = 'Geo::ShapeFile::Shape::Index';
 
 sub new {
     my $proto = shift;
@@ -50,10 +56,10 @@ sub parse_shp {
 
     $self->{source} = $self->{shp_data} = shift;
 
-    $self->extract_ints('big', 'shp_record_number', 'shp_content_length');
-    $self->extract_ints('little', 'shp_shape_type');
+    $self->_extract_ints('big', 'shp_record_number', 'shp_content_length');
+    $self->_extract_ints('little', 'shp_shape_type');
 
-    my $parser = 'parse_shp_' . $self->type($self->{shp_shape_type});
+    my $parser = '_parse_shp_' . $self->type($self->{shp_shape_type});
 
     croak "Can't parse shape_type $self->{shp_shape_type}"
       if !$self->can($parser);
@@ -70,7 +76,7 @@ sub parse_shp {
     }
 }
 
-sub parse_shp_Null {
+sub _parse_shp_Null {
     my $self = shift;
 }
 
@@ -104,15 +110,15 @@ sub calculate_bounds {
     my $self = shift;
 
     my %bounds = $self->find_bounds($self->points);
-    foreach(keys %bounds) {
+    foreach (keys %bounds) {
         $self->{'shp_' . $_} = $bounds{$_};
     }
     return %bounds;
 }
 
-sub parse_shp_Point {
+sub _parse_shp_Point {
     my $self = shift;
-    $self->extract_doubles('shp_X', 'shp_Y');
+    $self->_extract_doubles('shp_X', 'shp_Y');
     $self->{shp_points} = [Geo::ShapeFile::Point->new(
         X => $self->{shp_X},
         Y => $self->{shp_Y},
@@ -127,11 +133,11 @@ sub parse_shp_Point {
 # Double        X       // X coordinate
 # Double        Y       // Y coordinate
 
-sub parse_shp_PolyLine {
+sub _parse_shp_PolyLine {
     my $self = shift;
 
-    $self->extract_bounds();
-    $self->extract_parts_and_points();
+    $self->_extract_bounds();
+    $self->_extract_parts_and_points();
 }
 #  PolyLine
 # Double[4]             Box         // Bounding Box
@@ -140,11 +146,11 @@ sub parse_shp_PolyLine {
 # Integer[NumParts]     Parts       // Index to first point in part
 # Point[NumPoints]      Points      // Points for all parts
 
-sub parse_shp_Polygon {
+sub _parse_shp_Polygon {
     my $self = shift;
 
-    $self->extract_bounds();
-    $self->extract_parts_and_points();
+    $self->_extract_bounds();
+    $self->_extract_parts_and_points();
 }
 #  Polygon
 # Double[4]          Box        // Bounding Box
@@ -153,23 +159,23 @@ sub parse_shp_Polygon {
 # Integer[NumParts]  Parts      // Index to First Point in Part
 # Point[NumPoints]   Points     // Points for All Parts
 
-sub parse_shp_MultiPoint {
+sub _parse_shp_MultiPoint {
     my $self = shift;
 
-    $self->extract_bounds();
-    $self->extract_ints('little', 'shp_num_points');
-    $self->extract_points($self->{shp_num_points}, 'shp_points');
+    $self->_extract_bounds();
+    $self->_extract_ints('little', 'shp_num_points');
+    $self->_extract_points($self->{shp_num_points}, 'shp_points');
 }
 #  MultiPoint
 # Double[4]          Box        // Bounding Box
 # Integer            NumPoints  // Number of Points
 # Point[NumPoints]   Points     // The points in the set
 
-sub parse_shp_PointZ {
+sub _parse_shp_PointZ {
     my $self = shift;
 
-    $self->parse_shp_Point();
-    $self->extract_doubles('shp_Z', 'shp_M');
+    $self->_parse_shp_Point();
+    $self->_extract_doubles('shp_Z', 'shp_M');
     $self->{shp_points}->[0]->Z($self->{shp_Z});
     $self->{shp_points}->[0]->M($self->{shp_M});
 }
@@ -178,12 +184,12 @@ sub parse_shp_PointZ {
 # Double Z
 # Double M
 
-sub parse_shp_PolyLineZ {
+sub _parse_shp_PolyLineZ {
     my $self = shift;
 
-    $self->parse_shp_PolyLine();
-    $self->extract_z_data();
-    $self->extract_m_data();
+    $self->_parse_shp_PolyLine();
+    $self->_extract_z_data();
+    $self->_extract_m_data();
 }
 #  PolyLineZ
 # PolyLine +
@@ -192,12 +198,12 @@ sub parse_shp_PolyLineZ {
 # Double[2]             M Range
 # Double[NumPoints]     M Array
 
-sub parse_shp_PolygonZ {
+sub _parse_shp_PolygonZ {
     my $self = shift;
 
-    $self->parse_shp_Polygon();
-    $self->extract_z_data();
-    $self->extract_m_data();
+    $self->_parse_shp_Polygon();
+    $self->_extract_z_data();
+    $self->_extract_m_data();
 }
 #  PolygonZ
 # Polygon +
@@ -206,12 +212,12 @@ sub parse_shp_PolygonZ {
 # Double[2]             M Range
 # Double[NumPoints]     M Array
 
-sub parse_shp_MultiPointZ {
+sub _parse_shp_MultiPointZ {
     my $self = shift;
 
-    $self->parse_shp_MultiPoint();
-    $self->extract_z_data();
-    $self->extract_m_data();
+    $self->_parse_shp_MultiPoint();
+    $self->_extract_z_data();
+    $self->_extract_m_data();
 }
 #  MultiPointZ
 # MultiPoint +
@@ -220,57 +226,57 @@ sub parse_shp_MultiPointZ {
 # Double[2]         M Range
 # Double[NumPoints] M Array
 
-sub parse_shp_PointM {
+sub _parse_shp_PointM {
     my $self = shift;
 
-    $self->parse_shp_Point();
-    $self->extract_doubles('shp_M');
+    $self->_parse_shp_Point();
+    $self->_extract_doubles('shp_M');
     $self->{shp_points}->[0]->M($self->{shp_M});
 }
 #  PointM
 # Point +
 # Double M // M coordinate
 
-sub parse_shp_PolyLineM {
+sub _parse_shp_PolyLineM {
     my $self = shift;
 
-    $self->parse_shp_PolyLine();
-    $self->extract_m_data();
+    $self->_parse_shp_PolyLine();
+    $self->_extract_m_data();
 }
 #  PolyLineM
 # PolyLine +
 # Double[2]             MRange      // Bounding measure range
 # Double[NumPoints]     MArray      // Measures for all points
 
-sub parse_shp_PolygonM {
+sub _parse_shp_PolygonM {
     my $self = shift;
 
-    $self->parse_shp_Polygon();
-    $self->extract_m_data();
+    $self->_parse_shp_Polygon();
+    $self->_extract_m_data();
 }
 #  PolygonM
 # Polygon +
 # Double[2]             MRange      // Bounding Measure Range
 # Double[NumPoints]     MArray      // Measures for all points
 
-sub parse_shp_MultiPointM {
+sub _parse_shp_MultiPointM {
     my $self = shift;
 
-    $self->parse_shp_MultiPoint();
-    $self->extract_m_datextract_m_data();
+    $self->_parse_shp_MultiPoint();
+    $self->_extract_m_datextract_m_data();
 }
 #  MultiPointM
 # MultiPoint
 # Double[2]         MRange      // Bounding measure range
 # Double[NumPoints] MArray      // Measures
 
-sub parse_shp_MultiPatch {
+sub _parse_shp_MultiPatch {
     my $self = shift;
 
-    $self->extract_bounds();
-    $self->extract_parts_and_points();
-    $self->extract_z_data();
-    $self->extract_m_data();
+    $self->_extract_bounds();
+    $self->_extract_parts_and_points();
+    $self->_extract_z_data();
+    $self->_extract_m_data();
 }
 # MultiPatch
 # Double[4]           BoundingBox
@@ -284,23 +290,23 @@ sub parse_shp_MultiPatch {
 # Double[2]           M Range
 # Double[NumPoints]   M Array
 
-sub extract_bounds {
+sub _extract_bounds {
     my $self = shift;
 
-    $self->extract_doubles(qw/shp_x_min shp_y_min shp_x_max shp_y_max/);
+    $self->_extract_doubles(qw/shp_x_min shp_y_min shp_x_max shp_y_max/);
 }
 
-sub extract_ints {
+sub _extract_ints {
     my $self = shift;
     my $end = shift;
     my @what = @_;
 
     my $template = ($end =~ /^l/i) ? 'V' :'N';
 
-    $self->extract_and_unpack(4, $template, @what);
+    $self->_extract_and_unpack(4, $template, @what);
 }
 
-sub extract_count_ints {
+sub _extract_count_ints {
     my $self = shift;
     my $count = shift;
     my $end = shift;
@@ -308,14 +314,14 @@ sub extract_count_ints {
 
     my $template = ($end =~ /^l/i) ? 'V' :'N';
 
-    my $tmp = substr $self->{shp_data}, 0, ($count*4) ,'';
+    my $tmp = substr $self->{shp_data}, 0, ($count * 4), '';
     my @tmp = unpack $template . $count, $tmp;
     #my @tmp = unpack($template."[$count]",$tmp);
 
     $self->{$label} = [@tmp];
 }
 
-sub extract_doubles {
+sub _extract_doubles {
     my $self = shift;
     my @what = @_;
     my $size = 8;
@@ -323,33 +329,33 @@ sub extract_doubles {
 
     foreach ( @what ) {
         my $tmp = substr $self->{shp_data}, 0, $size, '';
-        $self->{ $_ } = $is_little_endian
+        $self->{ $_ } = $little_endian_sys
             ? (unpack $template, $tmp )
             : (unpack $template, scalar reverse $tmp );
     }
 }
 
-sub extract_count_doubles {
+sub _extract_count_doubles {
     my $self = shift;
     my $count = shift;
     my $label = shift;
 
     my $tmp = substr $self->{shp_data}, 0, $count*8, '';
-    my @tmp = $is_little_endian
+    my @tmp = $little_endian_sys
         ? (unpack 'd'.$count, $tmp )
         : (reverse unpack( 'd' . $count, scalar ( reverse( $tmp ) ) ) );
 
     $self->{$label} = [@tmp];
 }
 
-sub extract_points {
+sub _extract_points {
     my $self = shift;
     my $count = shift;
     my $label = shift;
 
     my $data = substr $self->{shp_data}, 0, $count * 16, '';
 
-    my @ps = $is_little_endian
+    my @ps = $little_endian_sys
         ? (unpack 'd*', $data )
         : (reverse unpack 'd*', scalar reverse $data );
 
@@ -361,7 +367,7 @@ sub extract_points {
     $self->{$label} = [@p];
 }
 
-sub extract_and_unpack {
+sub _extract_and_unpack {
     my $self = shift;
     my $size = shift;
     my $template = shift;
@@ -398,18 +404,25 @@ sub get_part {
     my $self  = shift;
     my $index = shift;
 
+    croak 'index passed to get_part must be >0'
+      if $index <= 0;
+
     $index -= 1; # shift to a 0 index
 
-    my @parts  = $self->parts;
-    my @points = $self->points;
-    my $beg    = $parts[$index]   || 0;
-    my $end    = $parts[$index+1] || 0;  #  if we use 5.010 then we can use the // operator here
+    #  $parts is an array of starting indexes in the $points array
+    my $parts  = $self->parts;
+    croak 'index exceeds number of parts'
+      if $index > $#$parts;
+
+    my $points = $self->points;
+    my $beg    = $parts->[$index]   || 0;
+    my $end    = $parts->[$index+1] || 0;  #  if we use 5.010 then we can use the // operator here
     $end -= 1;
     if ($end < 0) {
-        $end = $#points;
+        $end = $#$points;
     }
 
-    return @points[$beg .. $end];
+    return wantarray ? @$points[$beg .. $end] : [@$points[$beg .. $end]];
 }
 
 sub shape_type {
@@ -422,34 +435,34 @@ sub shape_id {
     return $self->{shp_record_number};
 }
 
-sub extract_z_data {
+sub _extract_z_data {
     my $self = shift;
 
-    $self->extract_doubles('shp_z_min', 'shp_z_max');
-    $self->extract_count_doubles($self->{shp_num_points}, 'shp_z_data');
+    $self->_extract_doubles('shp_z_min', 'shp_z_max');
+    $self->_extract_count_doubles($self->{shp_num_points}, 'shp_z_data');
     my @zdata = @{delete $self->{shp_z_data}};
     for (0 .. $#zdata) {
         $self->{shp_points}->[$_]->Z($zdata[$_]);
     }
 }
 
-sub extract_m_data {
+sub _extract_m_data {
     my $self = shift;
 
-    $self->extract_doubles ('shp_m_min','shp_m_max');
-    $self->extract_count_doubles($self->{shp_num_points}, 'shp_m_data');
+    $self->_extract_doubles ('shp_m_min', 'shp_m_max');
+    $self->_extract_count_doubles($self->{shp_num_points}, 'shp_m_data');
     my @mdata = @{delete $self->{shp_m_data}};
     for (0 .. $#mdata) {
         $self->{shp_points}->[$_]->M($mdata[$_]);
     }
 }
 
-sub extract_parts_and_points {
+sub _extract_parts_and_points {
     my $self = shift;
 
-    $self->extract_ints('little','shp_num_parts','shp_num_points');
-    $self->extract_count_ints($self->{shp_num_parts},'little','shp_parts');
-    $self->extract_points($self->{shp_num_points},'shp_points');
+    $self->_extract_ints('little', 'shp_num_parts', 'shp_num_points');
+    $self->_extract_count_ints($self->{shp_num_parts}, 'little', 'shp_parts');
+    $self->_extract_points($self->{shp_num_points}, 'shp_points');
 }
 
 
@@ -463,58 +476,250 @@ sub z_max { shift()->{shp_z_max}; }
 sub m_min { shift()->{shp_m_min}; }
 sub m_max { shift()->{shp_m_max}; }
 
+sub bounds {
+    my $self = shift;
+
+    my @results = (
+        $self->x_min,
+        $self->y_min,
+        $self->x_max,
+        $self->y_max,
+    );
+
+    return wantarray ? @results : \@results;
+}
+
 sub has_point {
     my $self  = shift;
     my $point = shift;
 
     return 0 if !$self->bounds_contains_point($point);
 
-    foreach ($self->points) {
-        return 1 if $_ == $point;
+    foreach my $check_pt ($self->points) {
+        return 1 if $check_pt == $point;
     }
 
     return 0;
 }
 
 sub contains_point {
-    my ( $self, $point ) = @_;
+    my ( $self, $point, $index_res ) = @_;
+
+    return $self->_contains_point_use_index ($point, $index_res)
+      if $self->get_spatial_index || defined $index_res;
 
     return 0 if !$self->bounds_contains_point( $point );
 
     my $a = 0;
-    my ( $x0, $y0 ) = ( $point->X, $point->Y );
+    my ( $x0, $y0 ) = ( $point->get_x, $point->get_y );
 
-    for ( 1 .. $self->num_parts ) {
-        my ( $x1, $y1 );
-        for my $p2 ( $self->get_part( $_ ) ) {
-            my $x2 = $p2->X - $x0;
-            my $y2 = $p2->Y - $y0;
+    #  one day we will track the bounds of the parts
+    #  so we can more easily skip parts of multipart polygons
+    my $num_parts = $self->num_parts;
 
-            if ( defined( $y1 ) && ( ( $y2 >= 0 ) != ( $y1 >= 0 ) ) ) {
-                my $isl = $x1*$y2 - $y1*$x2;
+    #  $x1, $x2, $y1 and $y2 are offsets from the point we are checking
+    foreach my $part_num (1 .. $num_parts) {
+        my $points = $self->get_part( $part_num );
+
+        my $p_start = shift @$points;  #  $points is a copy, so no harm in shifting
+        my $x1 = $p_start->get_x - $x0;
+        my $y1 = $p_start->get_y - $y0;
+
+        foreach my $p2 ( @$points ) {
+            my $x2 = $p2->get_x - $x0;
+            my $y2 = $p2->get_y - $y0;
+
+            #  does the ray intersect the segment?
+            if (($y2 >= 0) != ($y1 >= 0)) {
+                my $isl = $x1 * $y2 - $y1 * $x2;
                 if ( $y2 > $y1 ) {
-                    --$a if $isl > 0;
+                    if ($isl > 0) {
+                        $a--;
+                    }
                 }
                 else {
-                    ++$a if $isl < 0;
+                    if ($isl < 0) {
+                        $a++;
+                    }
                 }
             }
             ( $x1, $y1 ) = ( $x2, $y2 );
         }
     }
+
     return $a;
 }
 
+sub _contains_point_use_index {
+    my ( $self, $point, $index_res ) = @_;
+
+    return 0 if !$self->bounds_contains_point( $point );
+
+    my $sp_index_hash = $self->get_spatial_index || $self->build_spatial_index ($index_res);
+
+    my $a = 0;
+    my ( $x0, $y0 ) = ( $point->get_x, $point->get_y );
+
+    my @parts = $self->parts;
+    my $num_parts = scalar @parts;
+
+    #  $x1, $x2, $y1 and $y2 are offsets from the point we are checking
+  PART:
+    foreach my $part_index (1 .. $num_parts) {
+        my $sp_index = $sp_index_hash->{$part_index};
+
+        my @results = $sp_index->query_point($x0, $y0);
+
+        #  skip if not in this part's bounding box
+        next PART if !scalar @results;
+
+        # segments spanning the index's bounding box
+        for my $segment (@results) {
+
+            #  index stores bare x and y coords to avoid method overhead here
+            my $x1 = $segment->[0][0] - $x0;
+            my $y1 = $segment->[0][1] - $y0;
+            my $x2 = $segment->[1][0] - $x0;
+            my $y2 = $segment->[1][1] - $y0;
+
+            #  does the ray intersect the segment?
+            if (($y2 >= 0) != ($y1 >= 0)) {
+                my $isl = $x1 * $y2 - $y1 * $x2;
+                if ( $y2 > $y1 ) {
+                    if ($isl > 0) {
+                        $a--;
+                    }
+                }
+                else {
+                    if ($isl < 0) {
+                        $a++;
+                    }
+                }
+            }
+        }
+    }
+
+    return $a;
+}
+
+
+#  We could trigger a build if undefined,
+#  but save that for later.  
+sub get_spatial_index {
+    my $self = shift;
+
+    return $self->{_spatial_indexes};
+}
+
+#  Add the polygon's segments to a spatial index
+#  where the index boxes are as wide as the part
+#  they are in.
+#  The set of spatial indexes is a hash keyed by
+#  the part ID.
+#  $n is the number of boxes - need an automatic way of calculating, poss f(y_range / x_range)
+sub build_spatial_index {
+    my $self = shift;
+    my $n    = shift || 10;
+
+    $n = int $n;
+
+    croak 'Cannot build spatial index with <1 boxes'
+      if $n < 1;
+
+    my %sp_indexes;
+
+    my @parts = $self->parts;
+
+    my ($x_min, $x_max, $y_min, $y_max);
+
+    my $part_id = 0;
+    foreach my $part (@parts) {
+        $part_id ++;  #  parts are indexed base 1
+
+        my $segments = $self->get_segments ($part_id);
+
+        if (@parts > 1) {
+            my %bounds = $self->_get_part_bounds ($part_id);
+            ($x_min, $x_max, $y_min, $y_max) = @bounds{qw /x_min x_max y_min y_max/};
+        }
+        else {
+            ($x_min, $x_max, $y_min, $y_max) = $self->bounds;  #  faster than searching all points
+        }
+
+        my $n_boxes = @$segments > 20 ? $n : 1;
+        my $sp_index = $index_class->new ($n_boxes, $x_min, $y_min, $x_max, $y_max);
+
+        foreach my $segment (@$segments) {
+            my $p1 = $segment->[0];
+            my $p2 = $segment->[1];
+            my $y0 = $p1->get_y;
+            my $y1 = $p2->get_y;
+
+            #  reverse them if needed
+            if ($y1 < $y0) {
+                ($y0, $y1) = ($y1, $y0);
+            }
+
+            #  bare metal version
+            my $coords = [
+                [$p1->get_x, $p1->get_y],
+                [$p2->get_x, $p2->get_y],
+            ];
+
+            my @bbox = ($x_min, $y_min, $x_max, $y_max);
+            $sp_index->insert($coords, @bbox);
+        }
+
+        $sp_indexes{$part_id} = $sp_index;
+    }
+
+    $self->{_spatial_indexes} = \%sp_indexes;
+
+    return wantarray ? %sp_indexes : \%sp_indexes;
+}
+
+sub _get_part_bounds {
+    my $self = shift;
+    my $part = shift;
+
+    my $points = $self->get_part($part);
+
+    my $pt1 = shift @$points;
+    my ($x_min, $y_min) = ($pt1->get_x, $pt1->get_y);
+    my ($x_max, $y_max) = ($x_min, $y_min);
+
+    foreach my $pt (@$points) {
+        my $x = $pt->get_x;
+        my $y = $pt->get_y;
+
+        $x_min = min ($x_min, $x);
+        $y_min = min ($y_min, $y);
+        $x_max = max ($x_max, $x);
+        $y_max = max ($y_max, $y);
+    }
+
+    my %bounds = (
+        x_min => $x_min,
+        x_max => $x_max,
+        y_min => $y_min,
+        y_max => $y_max,
+    );
+
+    return wantarray ? %bounds : \%bounds;
+}
+
 sub get_segments {
     my $self = shift;
     my $part = shift;
 
-    my @points = $self->get_part($part);
-    my @segments = ();
-    for (0 .. $#points-1) {
-        push @segments, [$points[$_], $points[$_+1]];
+    my $points = $self->get_part($part);
+
+    my @segments;
+    foreach my $i (0 .. $#$points - 1) {
+        push @segments, [$points->[$i], $points->[$i+1]];
     }
-    return @segments;
+
+    return wantarray ? @segments : \@segments;
 }
 
 sub vertex_centroid {
@@ -592,7 +797,7 @@ sub dump {
 
     #$self->points();
     #$self->get_part();
-    #$self->x_min,x_max,y_min,y_max,z_min,z_max,m_min,m_max
+    #$self->x_min, x_max, y_min, y_max, z_min, z_max, m_min, m_max
 
     $return .= sprintf
         "Shape Type: %s (id: %d)  Parts: %d   Points: %d\n",
@@ -652,7 +857,7 @@ Geo::ShapeFile::Shape - Geo::ShapeFile utility class.
 
 =head1 ABSTRACT
 
-  This is a utility class for L<Geo::ShapeFile> that represents shapes.
+  This is a utility class for Geo::ShapeFile that represents shapes.
 
 =head1 DESCRIPTION
 
@@ -727,15 +932,34 @@ shape.  See the L<Geo::ShapeFile::Point> documentation for a note about how
 to exclude Z and/or M data from being considered when matching points.
 
 =item contains_point($point);
+=item contains_point($point, $use_index);
 
 Returns true if the specified point falls in the interior of this shape
 and false if the point is outside the shape.  Return value is unspecified
 if the point is one of the vertices or lies on some segment of the
 bounding polygon.
 
-Note that the return value is actually a winding-number computed ignoring
-Z and M fields and so will be negative if the point is contained within a
-shape winding the wrong way.
+Passing $use_index uses a spatial index if defined (building it if needed).
+See L<build_spatial_index> for more details. 
+This will be the default behaviour in a future release.
+
+Note that the algorithm uses a sidedness algorithm ignoring
+Z and M fields and so will likely not work if the point is contained within a
+shape winding the wrong way.  Polygon shapes should be anticlockwise for outer boundaries,
+and clockwise for inner void polygons.
+
+=item build_spatial_index ($index_res)
+Builds a spatial index for use in contains_point().
+$index_res is a positive integer which sets the nnumber of along the y-axis.
+A value of 0 lets the system determine the number.
+
+=item get_spatial_index()
+Gets the spatial index.  This is a hash indexed by part number.
+Returns a hash reference in scalar context.
+
+=item bounds
+Returns the object's bounds as an array (x_min, y_min, x_max, y_max).
+Returns an array ref in scalar context.
 
 =item get_segments($part)
 
