@@ -1,12 +1,11 @@
 package Geo::ShapeFile::Shape::Index;
-#use 5.010;  #  not yet
 use strict;
 use warnings;
 use POSIX qw /floor/;
 use Carp;
 use autovivification;
 
-our $VERSION = '2.99_002';
+our $VERSION = '2.56';
 
 #  should also handle X cells
 sub new {
@@ -73,10 +72,7 @@ sub snap_to_index {
     my $y_min = $self->get_y_min;
     my $y_res = $self->get_y_res;
 
-    #  take the floor, but add a small tolerance to
-    #  avoid precision issues with snapping
-    my $partial = ($y - $y_min) / $y_res;
-    my $y_block = floor ($partial * 1.001);
+    my $y_block = floor (($y - $y_min) / $y_res);
 
     return wantarray ? (0, $y_block) : "0:$y_block";
 }
